import json

import click

from numerai.cli.constants import *
from numerai.cli.util import files


@click.command()
@click.option(
    '--example', '-e', type=click.Choice(EXAMPLES), default=DEFAULT_EXAMPLE,
    help=f'Specify the example to copy, defaults to {DEFAULT_EXAMPLE}. '
         f'Options are {EXAMPLES}.'
)
@click.option(
    '--dest', '-d', type=str,
    help=f'Destination folder to which example code is written. '
         f'Defaults to the name of the example.'
)
@click.option('--verbose', '-v', is_flag=True)
def copy_example(example, dest, verbose):
    """
    Copies all example files into the current directory.

    WARNING: this will overwrite the following files if they exist:

        - Python: Dockerfile, model.py, train.py, predict.py, and requirements.txt

        - RLang:  Dockerfile, install_packages.R, main.R
    """
    files.copy_example(example, dest, verbose)


@click.command()
def list_constants():
    """
    Display default and constant values used by the CLI.

    Does NOT show currently configured node values.
    """
    click.secho(CONSTANTS_STR, fg='green')
    click.secho("SIZE_PRESETS:", fg='green')
    for size, preset in SIZE_PRESETS.items():
        suffix = '(default)' if size == DEFAULT_SIZE else ''
        click.secho(
            f'  {size} -> cpus: {preset[0] / 1024}, '
            f'mem: {preset[1] / 1024} GB {suffix}',
            fg='green' if size == DEFAULT_SIZE else 'yellow'
        )
    click.secho(
        "See https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-cpu-memory-error.html "
        "for more info about size presets.")



import os
import json
from pathlib import Path


TOURNAMENT_NUMERAI = 8
TOURNAMENT_SIGNALS = 11
PACKAGE_PATH = os.path.dirname(__file__)
CONFIG_PATH = os.path.join(str(Path.home()), '.numerai')
KEYS_PATH = os.path.join(CONFIG_PATH, '.keys')
NODES_PATH = os.path.join(CONFIG_PATH, 'nodes.json')
TERRAFORM_PATH = os.path.join(PACKAGE_PATH, "..", "terraform")
EXAMPLE_PATH = os.path.join(PACKAGE_PATH, "..", "examples")

EXAMPLES = os.listdir(EXAMPLE_PATH)

PROVIDER_AWS = "aws"
# PROVIDER_GCP = "gcp"
PROVIDERS = [
    PROVIDER_AWS,
    # PROVIDER_GCP
]

LOG_TYPE_WEBHOOK = 'webhook'
LOG_TYPE_CLUSTER = 'cluster'
LOG_TYPES = [
    LOG_TYPE_WEBHOOK,
    LOG_TYPE_CLUSTER
]

SIZE_PRESETS = {
    "gen-sm": (1024, 4096),
    "gen-md": (2048, 8192),
    "gen-lg": (4096, 16384),

    "cpu-xs": (512, 1024),
    "cpu-sm": (1024, 2048),
    "cpu-md": (2048, 4096),
    "cpu-lg": (4096, 8192),

    "mem-sm": (1024, 8192),
    "mem-md": (2048, 16384),
    "mem-lg": (4096, 30720),
}

DEFAULT_EXAMPLE = 'tournament-python3'
DEFAULT_SIZE = "mem-md"
DEFAULT_PROVIDER = PROVIDER_AWS
DEFAULT_PATH = os.getcwd()
DEFAULT_SETTINGS = {
    'provider': DEFAULT_PROVIDER,
    'cpu': SIZE_PRESETS[DEFAULT_SIZE][0],
    'memory': SIZE_PRESETS[DEFAULT_SIZE][1],
    'path': DEFAULT_PATH
}

CONSTANTS_STR = f'''Default values (not your configured node values):

---Tournament Numbers---
TOURNAMENT_NUMERAI: {TOURNAMENT_NUMERAI}
TOURNAMENT_SIGNALS: {TOURNAMENT_SIGNALS}

---Paths---
PACKAGE_PATH: {PACKAGE_PATH}
CONFIG_PATH: {CONFIG_PATH}
KEYS_PATH: {KEYS_PATH}
NODES_PATH: {NODES_PATH}
TERRAFORM_PATH: {TERRAFORM_PATH}
EXAMPLE_PATH: {EXAMPLE_PATH}

---Cloud Interaction---
PROVIDERS: {PROVIDERS}
LOG_TYPES: {LOG_TYPES}

---Prediction Nodes---
DEFAULT_EXAMPLE: {DEFAULT_EXAMPLE}
DEFAULT_SIZE: {DEFAULT_SIZE}
DEFAULT_PROVIDER: {DEFAULT_PROVIDER}
DEFAULT_PATH: {DEFAULT_PATH}
DEFAULT_SETTINGS: {json.dumps(DEFAULT_SETTINGS, indent=2)}
'''



import sys
import base64
import subprocess
from queue import Queue, Empty
from threading import Thread

import boto3
import click

from numerai.cli.constants import *
from numerai.cli.util.debug import root_cause
from numerai.cli.util.keys import sanitize_message, get_aws_keys, load_or_init_keys


def check_for_dockerfile(path):
    dockerfile_path = os.path.join(path, 'Dockerfile')
    if not os.path.exists(dockerfile_path):
        click.secho(
            f"No Dockerfile found in {path}, please ensure this node "
            f"was created from an example or follows the Prediction Node Architecture. "
            f"Learn More:\nhttps://github.com/numerai/numerai-cli/wiki/Prediction-Nodes",
            fg='red'
        )
        exit(1)
    if Path.home() == dockerfile_path:
        click.secho(
            f"DO NOT PUT THE DOCKERFILE IN YOUR HOME PATH, please ensure this node "
            f"was created from an example or follows the Prediction Node Architecture. "
            f"Learn More:\nhttps://github.com/numerai/numerai-cli/wiki/Prediction-Nodes",
            fg='red'
        )
        exit(1)


def subprocess_log(stream, queue):
    for line in iter(stream.readline, b''):
        queue.put(line)
    stream.close()


def get_from_q(q, verbose, default=b'', prefix=''):
    try:
        res = q.get(block=False)
        if verbose and res:
            click.secho(f'{prefix} {res.decode()}')
        return res
    except Empty as e:
        return default


def execute(command, verbose, censor_substr=None):
    if verbose:
        click.echo('Running: ' + sanitize_message(command, censor_substr))

    on_posix = 'posix' in sys.builtin_module_names
    proc = subprocess.Popen(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=1,
        close_fds=on_posix
    )
    stdout_q = Queue()
    stderr_q = Queue()
    stdout_t = Thread(
        target=subprocess_log,
        args=(proc.stdout, stdout_q)
    )
    stderr_t = Thread(
        target=subprocess_log,
        args=(proc.stderr, stderr_q)
    )

    try:
        stdout_t.start()
        stderr_t.start()
        stdout = b''
        stderr = b''
        while proc.poll() is None:
            stdout += get_from_q(stdout_q, verbose)
            stderr += get_from_q(stderr_q, verbose)

        returncode = proc.poll()
        if returncode != 0:
            root_cause(stdout, stderr)
    finally:
        stdout_t.join()
        stderr_t.join()
        proc.kill()

    return stdout, stderr


def format_if_docker_toolbox(path, verbose):
    '''
    Helper function to format if the system is running docker toolbox + mingw.
    Paths need to be formatted like unix paths, and the drive letter lower-cased
    '''
    if 'DOCKER_TOOLBOX_INSTALL_PATH' in os.environ and 'MSYSTEM' in os.environ:
        # '//' found working on win8.1 docker quickstart terminal, previously just '/'
        new_path = ('//' + path[0].lower() + path[2:]).replace('\\', '/')
        if verbose:
            click.secho(f"formatted path for docker toolbox: {path} -> {new_path}")
        return new_path
    return path


def build_tf_cmd(tf_cmd, env_vars, inputs, version, verbose):
    cmd = f"docker run"
    if env_vars:
        cmd += ' '.join([f' -e "{key}={val}"' for key, val in env_vars.items()])
    cmd += f' --rm -it -v {format_if_docker_toolbox(CONFIG_PATH, verbose)}:/opt/plan'
    cmd += f' -w /opt/plan hashicorp/terraform:{version} {tf_cmd}'
    if inputs:
        cmd += ' '.join([f' -var="{key}={val}"' for key, val in inputs.items()])
    return cmd


def terraform(tf_cmd, verbose, env_vars=None, inputs=None, version='0.14.3'):
    cmd = build_tf_cmd(tf_cmd, env_vars, inputs, version, verbose)
    stdout, stderr = execute(cmd, verbose)
    # if user accidently deleted a resource, refresh terraform and try again
    if b'ResourceNotFoundException' in stdout or b'NoSuchEntity' in stdout:
        refresh = build_tf_cmd('refresh', env_vars, inputs, version, verbose)
        execute(refresh, verbose)
        stdout, stderr = execute(cmd, verbose)
    return stdout


def build(node_config, verbose):
    numerai_keys = load_or_init_keys()['numerai']

    node_path = node_config["path"]
    curr_path = os.path.abspath('.')
    if curr_path not in node_path:
        raise RuntimeError(
            f'Current directory invalid, you must run this command either from'
            f' "{node_path}" or a parent directory of that path.'
        )
    path = node_path.replace(curr_path, '.').replace('\\', '/')
    if verbose:
        click.secho(f'Using relative path to node: {path}')

    build_arg_str = ''
    for arg in numerai_keys:
        build_arg_str += f' --build-arg {arg}={numerai_keys[arg]}'
    build_arg_str += f' --build-arg MODEL_ID={node_config["model_id"]}'
    build_arg_str += f' --build-arg SRC_PATH={path}'

    cmd = f'docker build --platform=linux/amd64 -t {node_config["docker_repo"]}' \
          f'{build_arg_str} -f {path}/Dockerfile .'

    execute(cmd, verbose)


def run(node_config, verbose, command=''):
    cmd = f"docker run --rm -it {node_config['docker_repo']} {command}"
    execute(cmd, verbose)


def login(node_config, verbose):
    if node_config['provider'] == PROVIDER_AWS:
        username, password = login_aws()

    else:
        raise ValueError(f"Unsupported provider: '{node_config['provider']}'")

    if os.name == 'nt':
        echo_cmd = f'echo | set /p="{password}"'
    else:
        echo_cmd = f'echo "{password}"'

    cmd = (
        echo_cmd +
        f" | docker login"
        f" -u {username}"
        f" --password-stdin"
        f" {node_config['docker_repo']}"
    )

    execute(cmd, verbose, censor_substr=password)


def login_aws():
    aws_public, aws_secret = get_aws_keys()
    ecr_client = boto3.client(
        'ecr', region_name='us-east-1',
        aws_access_key_id=aws_public,
        aws_secret_access_key=aws_secret)

    token = ecr_client.get_authorization_token()  # TODO: use registryIds
    username, password = base64.b64decode(
        token['authorizationData'][0]['authorizationToken']
    ).decode().split(':')

    return username, password


def push(docker_repo, verbose):
    cmd = f'docker push {docker_repo}'
    execute(cmd, verbose=verbose)


def cleanup(node_config):
    provider = node_config['provider']
    if provider == PROVIDER_AWS:
        imageIds = cleanup_aws(node_config['docker_repo'])

    else:
        raise ValueError(f"Unsupported provider: '{provider}'")

    if len(imageIds) > 0:
        click.secho(f"Deleted {str(len(imageIds))} old image(s) from remote docker repo", fg='yellow')


def cleanup_aws(docker_repo):
    aws_public, aws_secret = get_aws_keys()
    ecr_client = boto3.client(
        'ecr', region_name='us-east-1',
        aws_access_key_id=aws_public,
        aws_secret_access_key=aws_secret)

    docker_repo_name = docker_repo.split("/")[-1]

    resp = ecr_client.list_images(
        repositoryName=docker_repo_name,
        filter={'tagStatus': 'UNTAGGED'})

    imageIds = resp['imageIds']
    if len(imageIds) == 0:
        return []

    resp = ecr_client.batch_delete_image(
        repositoryName=docker_repo_name,
        imageIds=imageIds)

    return resp['imageIds']



import json
from configparser import \
    ConfigParser, \
    MissingSectionHeaderError

import boto3
import click
import numerapi

from numerai.cli.constants import *
from numerai.cli.constants import KEYS_PATH
from numerai.cli.util.debug import exception_with_msg
from numerai.cli.util.files import \
    load_or_init_nodes, \
    store_config, \
    maybe_create, \
    load_config


def reformat_keys():
    # REFORMAT OLD KEYS
    try:
        config = ConfigParser()
        config.read(KEYS_PATH)
        click.secho(f"Old keyfile format found, reformatting...", fg='yellow')

        new_config = {
            'aws': {
                'AWS_ACCESS_KEY_ID': config['default']['AWS_ACCESS_KEY_ID'],
                'AWS_SECRET_ACCESS_KEY': config['default']['AWS_SECRET_ACCESS_KEY']
            },
            'numerai': {
                'NUMERAI_PUBLIC_ID': config['default']['NUMERAI_PUBLIC_ID'],
                'NUMERAI_SECRET_KEY': config['default']['NUMERAI_SECRET_KEY']
            }
        }

        del config['default']
        with open(os.open(KEYS_PATH, os.O_CREAT | os.O_WRONLY, 0o600), 'w') as f:
            config.write(f)
            json.dump(new_config, f, indent=2)

    # if this file is already a json file skip
    except MissingSectionHeaderError:
        pass


def load_or_init_keys(provider=None):
    maybe_create(KEYS_PATH, protected=True)
    try:
        cfg = load_config(KEYS_PATH)
    except json.decoder.JSONDecodeError as e:
        reformat_keys()
        cfg = load_config(KEYS_PATH)
    if provider:
        return cfg[provider]
    return cfg


def get_numerai_keys():
    keys = load_or_init_keys()
    try:
        return keys['numerai']['NUMERAI_PUBLIC_ID'],\
               keys['numerai']['NUMERAI_SECRET_KEY']
    except KeyError:
        return None, None


def prompt_for_key(name, default):
    hidden = sanitize_message(default)
    new = click.prompt(name, default=hidden).strip()
    if new == hidden:
        return default
    return new


def config_numerai_keys():
    numerai_public, numerai_secret = get_numerai_keys()

    numerai_public = prompt_for_key('NUMERAI_PUBLIC_ID', numerai_public)
    numerai_secret = prompt_for_key('NUMERAI_SECRET_KEY', numerai_secret)
    check_numerai_validity(numerai_public, numerai_secret)

    keys_config = load_or_init_keys()
    keys_config.setdefault('numerai', {})
    keys_config['numerai']['NUMERAI_PUBLIC_ID'] = numerai_public
    keys_config['numerai']['NUMERAI_SECRET_KEY'] = numerai_secret
    store_config(KEYS_PATH, keys_config)


def check_numerai_validity(key_id, secret):
    try:
        napi = numerapi.NumerAPI(key_id, secret)
        napi.get_account()
    except Exception:
        raise exception_with_msg("Numerai keys seem to be invalid. "
                                 "Make sure you've entered them correctly.")


def get_provider_keys(node):
    provider = load_or_init_nodes(node)['provider']
    return load_or_init_keys(provider)


def get_aws_keys():
    keys = load_or_init_keys()
    try:
        return keys['aws']['AWS_ACCESS_KEY_ID'],\
               keys['aws']['AWS_SECRET_ACCESS_KEY']
    except KeyError:
        return None, None


def config_aws_keys():
    aws_public, aws_secret = get_aws_keys()
    aws_public = prompt_for_key('AWS_ACCESS_KEY_ID', aws_public)
    aws_secret = prompt_for_key('AWS_SECRET_ACCESS_KEY', aws_secret)
    check_aws_validity(aws_public, aws_secret)

    keys_config = load_or_init_keys()
    keys_config.setdefault('aws', {})
    keys_config['aws']['AWS_ACCESS_KEY_ID'] = aws_public
    keys_config['aws']['AWS_SECRET_ACCESS_KEY'] = aws_secret
    store_config(KEYS_PATH, keys_config)


def check_aws_validity(key_id, secret):
    try:
        client = boto3.client('s3', aws_access_key_id=key_id, aws_secret_access_key=secret)
        client.list_buckets()
    except Exception as e:
        if 'NotSignedUp' in str(e):
            raise exception_with_msg(
                f"Your AWS keys are valid, but the account is not finished signing up. "
                f"You either need to update your credit card in AWS at "
                f"https://portal.aws.amazon.com/billing/signup?type=resubscribe#/resubscribed, "
                f"or wait up to 24 hours for their verification process to complete."
            )

        raise exception_with_msg(
            f"AWS keys seem to be invalid. Make sure you've entered them correctly "
            f"and that your user has the necessary permissions (for help, see "
            f"https://github.com/numerai/numerai-cli/wiki/Amazon-Web-Services)."
        )


def config_provider_keys(cloud_provider):
    if cloud_provider == PROVIDER_AWS:
        config_aws_keys()


def sanitize_message(message, censor_substr=None):
    if message is None:
        return None
    all_keys = get_aws_keys() + get_numerai_keys()
    for key in all_keys:
        if key:
            try:
                message = message.replace(key, f'...{key[-5:]}')
            except AttributeError:
                continue
    if censor_substr:
        message = message.replace(
            censor_substr,
            f'...{censor_substr[-5:]}'
        )
    return message



import click

from numerai.cli.constants import *
from numerai.cli.util.docker import terraform
from numerai.cli.util.files import \
    maybe_create, \
    copy_files
from numerai.cli.util.keys import \
    config_numerai_keys, \
    config_provider_keys


@click.command()
@click.option(
    '--provider', '-p', type=str, default=DEFAULT_PROVIDER,
    help=f"Initialize with this providers API keys. Defaults to {DEFAULT_PROVIDER}.")
@click.option('--verbose', '-v', is_flag=True)
def setup(provider, verbose):
    """
    Initializes cli and provider API keys.
    """
    # check for old format, tell user to run numerai upgrade first
    if os.path.isfile(CONFIG_PATH) or os.path.isdir('.numerai'):
        click.secho('It looks like you have an old configuration of numerai-cli,'
                    'run `numerai upgrade` first.')
        return

    # setup numerai keys
    click.secho("Initializing numerai keys "
                "(press enter to keep value in brackets)...", fg='yellow')
    maybe_create(KEYS_PATH, protected=True)
    config_numerai_keys()

    # setup provider keys
    click.secho(f"\nInitializing {provider} keys "
                f"(press enter to keep value in brackets)...", fg='yellow')
    config_provider_keys(provider)

    # copy tf files
    click.secho("copying terraform files...")
    copy_files(TERRAFORM_PATH, CONFIG_PATH, force=True, verbose=True)

    # terraform init
    click.secho("initializing terraform to provision cloud infrastructure...")
    terraform("init -upgrade", verbose)

    click.secho("Numerai API Keys setup and working", fg='green')
    click.secho(f"{provider} API Keys setup and working", fg='green')
    click.secho(f"Terraform files copied to {CONFIG_PATH}", fg='green')
    click.echo('succesfully initialized numerai-cli')


import subprocess
import shutil

import click
from numerapi import base_api

from numerai.cli.constants import *
from numerai.cli.util.keys import \
    load_or_init_keys, \
    load_or_init_nodes, \
    get_numerai_keys
from numerai.cli.util.docker import terraform
from numerai.cli.util.debug import root_cause


@click.command()
def uninstall():
    """
    Removes cloud resources, local config, and python package.
    """
    click.secho(
    '''DANGER WILL ROBINSON, This will:
    - Destroy all nodes in the cloud
    - Remove all docker images on your computer
    - Delete the .numerai configuration directory on your computer
    - Uninstall the numerai-cli python package
    - Leave Python and Docker installed on your computer
    ''', fg='red'
    )
    if not click.confirm('Are you absolutely sure you want to uninstall?'):
        return

    if os.path.exists(CONFIG_PATH):
        if len(os.listdir(CONFIG_PATH)) == 0:
            os.rmdir(CONFIG_PATH)

        else:
            napi = base_api.Api(*get_numerai_keys())

            node_config = load_or_init_nodes()
            click.secho('deregistering all webhooks...')
            for node, config in node_config.items():
                napi.set_submission_webhook(config['model_id'], None)

            click.secho('destroying cloud resources...')
            all_keys = load_or_init_keys()
            provider_keys = {}
            for provider in PROVIDERS:
                if provider in all_keys.keys():
                    provider_keys.update(all_keys[provider])
            terraform('destroy -auto-approve',
                      verbose=True, env_vars=provider_keys,
                      inputs={'node_config_file': 'nodes.json'})

            click.secho('cleaning up docker images...')
            subprocess.run('docker system prune -f -a --volumes', shell=True)
            try:
                shutil.rmtree(CONFIG_PATH)
            except PermissionError as e:
                root_cause('', str(e))

    click.secho('uninstalling python package...')
    res = subprocess.run(
        'pip3 uninstall numerai-cli -y',
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True
    )

    if res.returncode != 0:
        if b'PermissionError' in res.stderr:
            click.secho(
                'uninstall failed due to permissions, '
                'run "pip3 uninstall numerai-cli -y" manually '
                'to ensure the package was uninstalled',
                fg='red'
            )
        else:
            click.secho(f'Unexpected error occurred:\n{res.stderr}', fg='red')

    click.secho("All those moments will be lost in time, like tears in rain.", fg='red')



import logging

import pandas as pd
import numpy as np
# a fork of yfinance that implements retries nicely, see requirements.txt
import yfinance

# read in ticker map
TICKER_MAP = pd.read_csv(
    "https://numerai-signals-public-data.s3-us-west-2.amazonaws.com/"
    "signals_ticker_map_w_bbg.csv"
)


def map_tickers(tickers, from_ticker, to_ticker):
    # create two maps for bbg -> yahoo and yahoo -> bbg
    ticker_map = dict(zip(
        TICKER_MAP[from_ticker],
        TICKER_MAP[to_ticker]
    ))
    new_tickers = tickers.map(ticker_map).dropna()
    logging.info(f"Number of tickers in {from_ticker} -> {to_ticker} map: {len(ticker_map)}")
    return new_tickers


def download_yahoo_data(yfinance_tickers):
    # download data
    n = 1000  # chunk row size
    chunk_df = [
        yfinance_tickers.iloc[i : i + n]
        for i in range(0, len(yfinance_tickers), n)
    ]

    concat_dfs = []
    for df in chunk_df:
        try:
            # threads=True -> faster; tickers fail, script hangs
            # threads=False -> slower; more tickers will succeed
            temp_df = yfinance.download(
                df.str.cat(sep=" "),
                start="2005-12-01",
                threads=False
            )
            temp_df = temp_df["Adj Close"].stack().reset_index()
            concat_dfs.append(temp_df)
        except:
            pass
    full_data = pd.concat(concat_dfs)

    # properly position and clean raw data, after taking adjusted close only
    full_data.columns = ["date", "ticker", "price"]
    full_data.set_index("date", inplace=True)

    return full_data


def relative_strength_index(prices, interval=10):
    """
    Computes Relative Strength Index given a price series and lookback interval
    See more here https://www.investopedia.com/terms/r/rsi.asp
    """
    delta = prices.diff()

    # copy deltas, set losses to 0, get rolling avg
    gains = delta.copy()
    gains[gains < 0] = 0
    avg_gain = gains.rolling(interval).mean()

    # copy deltas, set gains to 0, get rolling avg
    losses = delta.copy()
    losses[losses > 0] = 0
    avg_loss = losses.rolling(interval).mean().abs()

    # calculate relative strength and it's index
    rs = avg_gain / avg_loss
    rsi = 100.0 - (100.0 / (1.0 + rs))
    return rsi


def get_rsi_feature_names(num_days):
    # define column names of features, target, and prediction
    feat_quintile_lag = [f"RSI_quintile_lag_{num}" for num in range(num_days + 1)]
    feat_rsi_diff = [f"RSI_diff_{num}" for num in range(num_days)]
    feat_rsi_diff_abs = [f"RSI_abs_diff_{num}" for num in range(num_days)]
    return feat_quintile_lag, feat_rsi_diff, feat_rsi_diff_abs


def generate_rsi_features(full_data, num_days):
    # add Relative Strength Index
    logging.info('generating RSI for each price...')
    ticker_groups = full_data.groupby("ticker")
    full_data["RSI"] = ticker_groups["price"].transform(
        lambda x: relative_strength_index(x)
    )

    # group by era (date)
    logging.info('grouping by dates...')
    date_groups = full_data.groupby(full_data.index)

    # create quintile labels within each era, useful for learning relative ranking
    logging.info('generating RSI quintiles...')
    full_data["RSI_quintile"] = date_groups["RSI"].transform(
        lambda group: pd.qcut(group, 5, labels=False, duplicates="drop")
    )
    full_data.dropna(inplace=True)

    feat_quintile_lag, feat_rsi_diff, feat_rsi_diff_abs = get_rsi_feature_names(num_days)

    # create lagged features grouped by ticker
    logging.info('grouping by ticker...')
    ticker_groups = full_data.groupby("ticker")

    # lag 0 is that day's value, lag 1 is yesterday's value, etc
    logging.info('generating lagged RSI quintiles...')
    for day in range(num_days + 1):
        full_data[feat_quintile_lag[day]] = ticker_groups["RSI_quintile"].transform(
            lambda group: group.shift(day)
        )

    # create difference of the lagged features and
    # absolute difference of the lagged features (change in RSI quintile by day)
    logging.info('generating lagged RSI diffs...')
    for day in range(num_days):
        full_data[feat_rsi_diff[day]] = (
            full_data[feat_quintile_lag[day]] - full_data[feat_quintile_lag[day + 1]]
        )
        full_data[feat_rsi_diff_abs[day]] = np.abs(full_data[feat_rsi_diff[day]])

    return full_data, feat_quintile_lag + feat_rsi_diff + feat_rsi_diff_abs


def add_targets_and_split(full_data):
    logging.info('adding targets...')
    # read in Signals targets and format the date
    targets = pd.read_csv(
        "https://numerai-signals-public-data.s3-us-west-2.amazonaws.com/"
        "signals_train_val_bbg.csv"
    )
    targets["date"] = pd.to_datetime(
        targets["friday_date"], format="%Y%m%d"
    ).dt.strftime("%Y-%m-%d")

    full_data.reset_index(inplace=True)
    full_data["date"] = full_data["date"].astype(str)

    # merge our feature data with Numerai targets
    logging.info('generating dataset...')
    ml_data = pd.merge(
        full_data, targets,
        on=["date", "bloomberg_ticker"]
    )

    # convert date to datetime and index on it
    ml_data["date"] = pd.to_datetime(targets['date'], format='%Y-%m-%d')
    ml_data = ml_data.set_index("date")

    # for training and testing we want clean, complete data only
    ml_data.dropna(inplace=True)
    # ensure we have only fridays
    ml_data = ml_data[ml_data.index.weekday == 4]
    # drop eras with under 50 observations per era
    ml_data = ml_data[ml_data.index.value_counts() > 50]

    # train test split
    train_data = ml_data[ml_data["data_type"] == "train"]
    test_data = ml_data[ml_data["data_type"] == "validation"]

    return train_data, test_data



import click
from colorama import init

from numerai.cli import \
    constants, \
    doctor, \
    node, \
    setup, \
    uninstall, \
    upgrade, \
    misc
from numerai.cli.util import \
    debug, \
    docker, \
    files, \
    keys


@click.group()
def numerai():
    """
    This tool helps you setup Numer.ai Prediction Nodes and deploy your models to them.
    """
    pass


def main():
    if debug.is_win8():
        init(wrap=False)
    else:
        init(autoreset=True)

    numerai.add_command(setup.setup)
    numerai.add_command(node.node)
    numerai.add_command(doctor.doctor)
    numerai.add_command(uninstall.uninstall)
    numerai.add_command(upgrade.upgrade)
    numerai.add_command(misc.copy_example)
    numerai.add_command(misc.list_constants)
    numerai()



import json
import subprocess
import sys
from urllib import request

import click

from numerai.cli.util.files import load_or_init_nodes
from numerai.cli.util.debug import is_win8, is_win10
from numerai.cli.util.keys import \
    check_aws_validity, \
    check_numerai_validity, \
    get_numerai_keys, \
    get_aws_keys


@click.command()
def doctor():
    """
    Checks and repairs your environment in case of errors.
    Attempts to provide information to debug your local machine.
    """
    # Check environment pre-reqs
    click.secho("Running the environment setup script for your OS...")
    env_setup_cmd = None
    env_setup_status = -1
    env_setup_err = ''
    if sys.platform == "linux" or sys.platform == "linux2":
        env_setup_cmd = 'sudo apt update && sudo apt install -y libcurl4 curl && ' \
                        'sudo curl https://raw.githubusercontent.com/numerai/numerai-cli/master/scripts/setup-ubu.sh ' \
                        '| sudo bash'

    elif sys.platform == "darwin":
        env_setup_cmd = 'curl https://raw.githubusercontent.com/numerai/numerai-cli/master/scripts/setup-mac.sh | bash'

    elif is_win10():
        env_setup_cmd = 'powershell -command "$Script = Invoke-WebRequest ' \
                        '\'https://raw.githubusercontent.com/numerai/numerai-cli/master/scripts/setup-win10.ps1\'; ' \
                        '$ScriptBlock = [ScriptBlock]::Create($Script.Content); Invoke-Command -ScriptBlock $ScriptBlock"'

    elif is_win8():
        # TODO: check if more is needed?
        env_setup_cmd = 'docker info'

    else:
        env_setup_status = 1
        env_setup_err = f"Unrecognized Operating System {sys.platform}, " \
                        f"cannot run environment setup script, skipping..."

    if env_setup_cmd is not None:
        res = subprocess.run(
            env_setup_cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        env_setup_status = res.returncode
        env_setup_err = res.stderr

    # Check official (non-dev) version
    click.secho(f"Checking your numerai-cli version...")
    res = str(subprocess.run(
        'pip3 show numerai-cli',
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True
    ))
    curr_ver = [s for s in res.split('\\n') if 'Version:' in s][0].split(': ')[1]
    url = f"https://pypi.org/pypi/numerai-cli/json"
    versions = list(reversed(sorted(filter(
        lambda key: 'dev' not in key,
        json.load(request.urlopen(url))["releases"].keys()
    ))))

    # Check keys
    click.secho("Checking your API keys...")
    nodes_config = load_or_init_nodes()
    used_providers = [nodes_config[n]['provider'] for n in nodes_config]

    invalid_providers = []
    try:
        check_numerai_validity(*get_numerai_keys())
    except:
        invalid_providers.append('numerai')
    if 'aws' in used_providers:
        try:
            check_aws_validity(*get_aws_keys())
        except:
            invalid_providers.append('aws')

    if env_setup_status != 0:
        click.secho(f"Environment setup incomplete:", fg='red')
        click.secho(env_setup_err, fg='red')
        click.secho(f"Ensure your OS is supported and read the Troubleshooting wiki: "
                    f"https://github.com/numerai/numerai-cli/wiki/Troubleshooting", fg='red')
    else:
        click.secho("Environment setup with Docker and Python", fg='green')

    if curr_ver < versions[0]:
        click.secho(f"numerai-cli needs an upgrade"
                    f"(run `pip3 install -U numerai-cli` to fix)", fg='red')
    else:
        click.secho("numerai-cli is up to date", fg='green')

    if len(invalid_providers):
        click.secho(f"Invalid provider keys: {invalid_providers}"
                    f"(run `numerai setup` to fix)", fg='red')

    else:
        click.secho("API Keys working", fg='green')

    click.secho(
        "\nIf you need help troubleshooting or want to report a bug please read the" 
        "\nTroubleshooting and Feedback section of the readme:"
        "\nhttps://github.com/numerai/numerai-cli#troubleshooting-and-feedback",
        fg='yellow'
    )



import os
import logging
from datetime import datetime, timedelta
import configparser

from data import \
    download_yahoo_data,\
    map_tickers,\
    generate_rsi_features,\
    add_targets_and_split, \
    get_rsi_feature_names

import joblib
import numerapi
import pandas as pd
from sklearn.ensemble import GradientBoostingRegressor
from dateutil.relativedelta import relativedelta, FR

TARGET_NAME = "target"
PREDICTION_NAME = "signal"
TRAINED_MODEL_PREFIX = './trained_model'

# Pull model id from "MODEL_ID" environment variable
# defaults to None, change to a model id from
MODEL_ID = os.getenv('MODEL_ID', None)
MODEL = GradientBoostingRegressor(subsample=0.1)

napi = numerapi.SignalsAPI()


def download_data(live_data_date):
    eligible_tickers = pd.Series(napi.ticker_universe(), name="bloomberg_ticker")
    logging.info(f"Number of eligible tickers: {len(eligible_tickers)}")

    yfinance_tickers = map_tickers(eligible_tickers, "bloomberg_ticker", "yahoo")
    logging.info(f"Number of yahoo tickers: {len(yfinance_tickers)}")

    num_days_lag = 5
    if os.path.exists('full_data.csv'):
        full_data = pd.read_csv('full_data.csv')
        quintile_lag, rsi_diff, rsi_diff_abs = get_rsi_feature_names(num_days_lag)
        feature_names = quintile_lag + rsi_diff + rsi_diff_abs

    else:
        full_data = download_yahoo_data(yfinance_tickers)

        full_data["bloomberg_ticker"] = map_tickers(
            full_data.ticker, "yahoo", "bloomberg_ticker")
        logging.info(
            f"Num tickers with data: {len(full_data.bloomberg_ticker.unique())}")
        logging.info(f"data size: {full_data.shape}")

        full_data, feature_names = generate_rsi_features(full_data, num_days=num_days_lag)
        logging.info(f"Features for training:\n {feature_names}")
        full_data.to_csv('full_data.csv')

    # add numerai targets and do train/test split
    train_data, test_data = add_targets_and_split(full_data)
    logging.info(f"{len(train_data)} train examples and {len(test_data)} test examples")

    # generate live data
    date_string = live_data_date.strftime("%Y-%m-%d")
    try:
        live_data = full_data[full_data.date == date_string].copy()
    except KeyError as e:
        logging.error(f"No ticker on {e}")
        live_data = full_data.iloc[:0].copy()
    live_data.dropna(subset=feature_names, inplace=True)

    # get data from the day before, for markets that were closed
    last_thursday = live_data_date - timedelta(days=1)
    thursday_date_string = last_thursday.strftime("%Y-%m-%d")
    thursday_data = full_data[full_data.date == thursday_date_string]

    # Only select tickers than aren't already present in live_data
    thursday_data = thursday_data[
        ~thursday_data.ticker.isin(live_data.ticker.values)
    ].copy()
    thursday_data.dropna(subset=feature_names, inplace=True)

    live_data = pd.concat([live_data, thursday_data])
    live_data = live_data.set_index("date")

    return feature_names, train_data, test_data, live_data


def train(train_data, feature_names, model_id, model, force_training=False):
    model_name = TRAINED_MODEL_PREFIX
    if model_id:
        model_name += f"_{model_id}"

    # load a model if we have a trained model already and we aren't forcing a training session
    if os.path.exists(model_name) and not force_training:
        logging.info('loading existing trained model')
        model = joblib.load(model_name)
        return model

    logging.info('training model')
    model.fit(train_data[feature_names], train_data[TARGET_NAME])
    del train_data

    logging.info('saving model')
    joblib.dump(model, model_name)

    return model


def predict(test_data, live_data, live_data_date, feature_names, model):
    # predict test data
    test_data[PREDICTION_NAME] = model.predict(test_data[feature_names])

    logging.info(f"Number of live tickers to submit: {len(live_data)}")
    live_data[PREDICTION_NAME] = model.predict(live_data[feature_names])

    # prepare and writeout example file
    diagnostic_df = pd.concat([test_data, live_data])
    del test_data
    del live_data

    diagnostic_df["friday_date"] = diagnostic_df.friday_date.fillna(
        live_data_date.strftime("%Y%m%d")
    ).astype(int)
    diagnostic_df["data_type"] = diagnostic_df.data_type.fillna("live")

    return diagnostic_df[[
        "bloomberg_ticker",
        "friday_date",
        "data_type",
        "signal"
    ]].reset_index(drop=True)


def submit(predictions, predict_output_path, model_id=None):
    logging.info('writing predictions to file')
    # numerai doesn't want the index, so don't write it to our file
    predictions.to_csv(predict_output_path, index=False)

    # Numerai API uses Environment variables NUMERAI_PUBLIC_ID and NUMERAI_SECRET_KEY
    # these are set by docker via the numerai cli; see README for more info
    logging.info(f'submitting for {model_id}')
    napi.upload_predictions(predict_output_path, model_id=model_id)


def main():
    """
    Creates example_signal_upload.csv to upload for validation and live data submission
    """
    # choose data as of most recent friday
    last_friday = datetime.now() + relativedelta(weekday=FR(-1))
    feature_names, train_data, test_data, live_data = download_data(last_friday)

    model = train(train_data, feature_names, MODEL_ID, MODEL)
    predictions = predict(test_data, live_data, last_friday, feature_names, model)
    submit(predictions, "example_signals.csv", MODEL_ID)


if __name__ == "__main__":
    main()


import click
from numerapi import base_api

from numerai.cli.constants import *
from numerai.cli.util.docker import terraform
from numerai.cli.util.files import \
    load_or_init_nodes, \
    store_config
from numerai.cli.util.keys import get_provider_keys, get_numerai_keys


@click.command()
@click.option('--verbose', '-v', is_flag=True)
@click.pass_context
def destroy(ctx, verbose):
    """
    Uses Terraform to destroy Numerai Compute cluster in AWS.
    This will delete everything, including:
        - lambda url
        - docker container and associated task
        - all logs
    This command is idempotent and safe to run multiple times.
    """
    ctx.ensure_object(dict)
    model = ctx.obj['model']
    node = model['name']
    if not os.path.exists(CONFIG_PATH):
        click.secho(f".numerai directory not setup, run `numerai setup`...", fg='red')
        return

    try:
        nodes_config = load_or_init_nodes()
        node_config = nodes_config[node]
        provider_keys = get_provider_keys(node)
    except (KeyError, FileNotFoundError) as e:
        click.secho(f"make sure you run `numerai setup` and "
                    f"`numerai node -n {node} config` first...", fg='red')
        return

    try:
        click.secho(f"deleting node configuration...")
        del nodes_config[node]
        store_config(NODES_PATH, nodes_config)

        click.secho(f"deleting cloud resources for node...")
        terraform(f'apply -auto-approve', verbose,
                  env_vars=provider_keys,
                  inputs={'node_config_file': 'nodes.json'})

    except Exception as e:
        click.secho(e.__str__(), fg='red')
        nodes_config[node] = node_config
        store_config(NODES_PATH, nodes_config)
        return

    if 'model_id' in node_config and 'webhook_url' in node_config:
        napi = base_api.Api(*get_numerai_keys())
        model_id = node_config['model_id']
        webhook_url = node_config['webhook_url']
        click.echo(f'deregistering webhook {webhook_url} for model {model_id}...')
        napi.set_submission_webhook(model_id, None)

    click.secho("Prediction Node destroyed successfully", fg='green')


import json
import shutil

import click

from numerai.cli.constants import *
from numerai.cli.util.docker import terraform
from numerai.cli.util.files import \
    copy_files
from numerai.cli.util.keys import \
    load_or_init_keys, \
    load_or_init_nodes, \
    config_numerai_keys, \
    config_provider_keys


@click.command()
@click.option('--verbose', '-v', is_flag=True)
def upgrade(verbose):
    """
    Upgrades configuration from <=0.2 format to >=0.3 format.
    """
    home = str(Path.home())
    old_key_path = os.path.join(home, '.numerai')
    old_config_path = os.path.join(os.getcwd(), '.numerai/')

    if str(old_key_path) == str(old_config_path):
        click.secho('You cannot run this command from your home directory.')
        return

    if not os.path.exists(old_config_path):
        click.secho(
            'Run this command from the directory in which you first ran `numerai setup`'
            ' for CLI version 0.1 and 0.2 (it should have the .numerai folder in it)'
        )
        return

    click.secho(f"Upgrading, do not interrupt or else "
                f"your environment may be corrupted.", fg='yellow')

    # MOVE KEYS FILE
    if os.path.isfile(old_key_path):
        temp_key_path = os.path.join(old_config_path, '.keys')
        click.secho(f"\tmoving old keyfile from '{old_key_path}' to '{temp_key_path}'",)
        shutil.move(old_key_path, temp_key_path)

    # MOVE CONFIG FILE
    if os.path.exists(old_config_path):
        click.secho(f"\tmoving old config from {old_config_path} to {CONFIG_PATH}",)
        shutil.move(old_config_path, CONFIG_PATH)

    # INIT KEYS AND NODES
    keys_config = load_or_init_keys()
    if not os.path.exists(KEYS_PATH) or 'aws' not in keys_config or 'numerai' not in keys_config:
        click.secho(f"Keys missing from {KEYS_PATH}, you must re-initialize your keys:")
        config_numerai_keys()
        config_provider_keys(PROVIDER_AWS)
    nodes_config = load_or_init_nodes()

    # DELETE OLD CONFIG FILES
    click.secho('Checking for old config output files...', fg='yellow')
    old_suburl_path = os.path.join(CONFIG_PATH, 'submission_url.txt')
    if os.path.exists(old_suburl_path):
        click.secho(f"\tdeleting {old_suburl_path}, you can populate the "
                    f"new nodes.json file with `numerai node config`")
        os.remove(old_suburl_path)
    old_docker_path = os.path.join(CONFIG_PATH, 'docker_repo.txt')
    if os.path.exists(old_docker_path):
        click.secho(f"\tdeleting {old_docker_path}, you can populate the "
                    f"new nodes.json file with `numerai node config`")
        os.remove(old_docker_path)

    # UPGRADE, RENAME, AND UPDATE TERRAFORM FILES
    click.secho('Upgrading terraform files...', fg='yellow')
    try:
        with open(os.path.join(CONFIG_PATH, 'terraform.tfstate')) as f:
            tfstate = json.load(f)
        keys_config = load_or_init_keys('aws')
        if '0.12' in tfstate['terraform_version']:
            terraform('0.13upgrade -yes', verbose, version='0.13.6', env_vars=keys_config)
            terraform('init', verbose, version='0.13.6', env_vars=keys_config)
            terraform('apply -auto-approve', verbose, version='0.13.6', env_vars=keys_config)
    except FileNotFoundError:
        pass
    except click.ClickException:
        click.secho("Failed to upgrade to terraform state!", fg='red')
        return
    except Exception as e:
        click.secho(f'Uncaught exception: {str(e)}', fg='red')
        return

    tf_files_map = {
        'main.tf': '-main.tf',
        'variables.tf': '-inputs.tf',
        'outputs.tf': '-outputs.tf'
    }
    for old_file, new_file in tf_files_map.items():
        old_file = os.path.join(CONFIG_PATH, old_file)
        new_file = os.path.join(CONFIG_PATH, new_file)
        if not os.path.exists(old_file):
            continue
        if not os.path.exists(old_file):
            click.secho(f'\trenaming {old_file} to {new_file} to prep for upgrade...')
            shutil.move(old_file, new_file)
        else:
            os.remove(old_file)
    copy_files(
        TERRAFORM_PATH,
        CONFIG_PATH,
        force=True,
        verbose=verbose
    )

    # terraform init
    click.secho("Re-initializing terraform...", fg='yellow')
    terraform("init -upgrade", verbose=verbose)

    if click.confirm("It's recommended you destroy your current Compute Node. Continue?"):
        click.secho("Removing old cloud infrastructure...", fg='yellow')
        terraform('destroy -auto-approve', verbose,
                  env_vars=load_or_init_keys('aws'),
                  inputs={'node_config_file': 'nodes.json'})

    click.secho('Upgrade complete!', fg='green')
    click.secho('run `numerai node config --help` to learn how to '
                'register this directory as a prediction node')



import os

import click

from numerai.cli.util import files, docker


@click.command()
@click.option('--verbose', '-v', is_flag=True)
@click.pass_context
def deploy(ctx, verbose):
    """Builds and pushes your docker image to the AWS ECR repo"""
    ctx.ensure_object(dict)
    model = ctx.obj['model']
    node = model['name']
    node_config = files.load_or_init_nodes(node)

    docker.check_for_dockerfile(node_config['path'])

    click.echo('building container image (this may take several minutes)...')
    docker.build(node_config, verbose)

    click.echo('logging into container registry...')
    docker.login(node_config, verbose)

    click.echo('pushing image to registry (this may take several minutes)...')
    docker.push(node_config['docker_repo'], verbose)

    click.echo('cleaning up local images...')
    docker.cleanup(node_config)

    click.secho('Prediction Node deployed. Next: test your node.', fg='green')



import time
import json
from datetime import datetime, timedelta

import boto3
import click
import requests
from numerapi import base_api

from numerai.cli.constants import *
from numerai.cli.util import docker
from numerai.cli.util.debug import exception_with_msg
from numerai.cli.util.files import load_or_init_nodes
from numerai.cli.util.keys import get_aws_keys, get_numerai_keys


@click.command()
@click.option(
    '--local', '-l', type=str, is_flag=True,
    help=f'Test the container locally, uses value specified with --command. ')
@click.option(
    '--command', '-c', type=str, default="",
    help=f'Used to override the terminal command during local testing. '
         f'Defaults to the command specified in the Dockerfile.')
@click.option('--verbose', '-v', is_flag=True)
@click.pass_context
def test(ctx, local, command, verbose):
    """
    Full end-to-end cloud or local test for a Prediction Node.

    This checks that:
    1. Numerai can reach the Trigger
    2. The Trigger schedules a Container to run
    3. The Container starts up on the Compute Cluster
    4. The Container uploads a submission with the Trigger ID assigned to it
    """
    ctx.ensure_object(dict)
    model = ctx.obj['model']
    node = model['name']
    is_signals = model['is_signals']
    node_config = load_or_init_nodes(node)

    if local:
        click.secho("starting local test; building container...")
        docker.build(node_config, verbose)
        click.secho("running container...")
        docker.run(node_config, verbose, command=command)

    api = base_api.Api(*get_numerai_keys())
    trigger_id = None
    try:
        if 'cron' in node_config:
            click.secho("Attempting to manually trigger Cron node...")
            res = requests.post(node_config['webhook_url'], json.dumps({}))
            res.raise_for_status()


        else:
            click.secho("Checking if Numerai can Trigger your model...")
            res = api.raw_query(
                '''mutation ( $modelId: String! ) {
                    triggerModelWebhook( modelId: $modelId )
                }''',
                variables={'modelId': node_config['model_id'],},
                authorization=True
            )
            trigger_id = res['data']['triggerModelWebhook']
            click.secho(f"Trigger ID assigned for this test: {trigger_id}", fg='green')

        if verbose:
            click.echo(f"response:\n{res}")
        click.secho(f"Webhook reachable...", fg='green')

    except ValueError as e:
        click.secho(f'there was a problem calling your webhook...', fg='red')
        if 'Internal Server Error' in str(e):
            click.secho('attempting to dump webhook logs', fg='red')
            monitor(node, node_config, True, 20, LOG_TYPE_WEBHOOK, False)
        return

    click.secho("checking task status...")
    monitor(node, node_config, verbose, 15, LOG_TYPE_CLUSTER, follow_tail=True)

    click.secho("checking for submission...")
    res = api.raw_query(
        '''query ( $modelId: String! ) {
            submissions( modelId: $modelId ){
                round{ number, tournament },
                triggerId
                insertedAt
            }
        }''',
        variables={'modelId': node_config['model_id']},
        authorization=True
    )
    tournament = TOURNAMENT_SIGNALS if is_signals else TOURNAMENT_NUMERAI
    curr_round = api.get_current_round(tournament)
    latest_subs = sorted(
        filter(
            lambda sub: sub['round']['number'] == curr_round,
            res['data']['submissions']
        ),
        key=lambda sub: sub['insertedAt'],
        reverse=True
    )
    if len(latest_subs) == 0:
        click.secho("No submission found for current round, test failed", fg='red')
        return

    latest_sub = latest_subs[0]

    if 'cron' in node_config:
        latest_date = datetime.strptime(latest_sub['insertedAt'], "%Y-%m-%dT%H:%M:%SZ")
        if latest_date < datetime.utcnow() - timedelta(minutes=5):
            click.secho(
                "No submission appeared in the last 5 minutes, be sure that your node"
                " is submitting correctly, check the numerai-cli wiki for more"
                " information on how to monitor parts of your node.",
                fg='red'
            )

    if trigger_id != latest_sub['triggerId']:
        click.secho(
            "Your node did not submit the Trigger ID assigned during this test, "
            "please ensure your node uses numerapi >= 0.2.4 (ignore if using rlang)",
            fg='red'
        )
        return

    click.secho("Submission uploaded correctly", fg='green')
    click.secho("Test complete, your model now submits automatically!", fg='green')


def monitor(node, config, verbose, num_lines, log_type, follow_tail):
    if log_type not in LOG_TYPES:
        raise exception_with_msg(f"Unknown log type '{log_type}', "
                                 f"must be one of {LOG_TYPES}")

    if config['provider'] == PROVIDER_AWS:
        monitor_aws(node, config, num_lines, log_type, follow_tail, verbose)

    else:
        click.secho(f"Unsupported provider: '{config['provider']}'", fg='red')
        return


def monitor_aws(node, config, num_lines, log_type, follow_tail, verbose):
    aws_public, aws_secret = get_aws_keys()
    logs_client = boto3.client(
        'logs', region_name='us-east-1',
        aws_access_key_id=aws_public,
        aws_secret_access_key=aws_secret)
    ecs_client = boto3.client(
        'ecs', region_name='us-east-1',
        aws_access_key_id=aws_public,
        aws_secret_access_key=aws_secret)

    if log_type == LOG_TYPE_WEBHOOK:
        get_name_and_print_logs(logs_client, config['api_log_group'], num_lines)
        get_name_and_print_logs(logs_client, config['webhook_log_group'], num_lines)
        return

    if log_type == LOG_TYPE_CLUSTER:
        family = config['cluster_log_group']

        # wait until log stream has been created
        i = 0
        name = None
        while name is None:
            i += 1
            task = get_recent_task_status_aws(ecs_client, node, verbose)
            if task is None:
                get_name_and_print_logs(logs_client, family, num_lines)
                return
            task_id = task["taskArn"].split('/')[-1]

            streams = logs_client.describe_log_streams(
                logGroupName=family,
                logStreamNamePrefix=f"ecs/{node}/{task_id}"
            )
            streams = list(filter(
                lambda s: s['logStreamName'].endswith(task_id),
                streams['logStreams']
            ))

            msg = f"Task status: {task['lastStatus']}."

            if task['lastStatus'] == "STOPPED":
                if len(streams) == 0:
                    click.secho(f"{msg} No log file, did you deploy?", fg='yellow')
                    exit(1)
                else:
                    click.secho(f"{msg} Checking for log events...", fg='green')
                    break

            elif len(streams) == 0:
                click.secho(
                    f"{msg} Waiting for log file to be created..."
                    f"{'.' * i}\r", fg='yellow', nl=False
                )
                time.sleep(2)

            else:
                name = streams[0]['logStreamName']
                click.secho(f"\n{msg} Log file created: {name}", fg='green')
                break

        # print out the logs
        next_token, num_events = print_logs(logs_client, family, name, limit=num_lines)
        total_events = num_events
        while follow_tail:
            next_token, num_events = print_logs(
                logs_client, family, name,
                next_token=(next_token if total_events > 0 else None)
            )
            total_events += num_events
            if total_events == 0:
                click.secho(f"Waiting for log events...\r", fg='yellow', nl=False)

            task = get_recent_task_status_aws(ecs_client, node, verbose)

            if task['lastStatus'] == "STOPPED":
                click.secho(f"\nTask is stopping...", fg='yellow')

                if len(task['containers']) and 'reason' in task['containers'][0]:
                    container = task['containers'][0]
                    click.secho(f"Container Exit code: {container['exitCode']}\n"
                                f"Reason: {container['reason']}", fg='red')
                break

        start = time.time()
        if total_events == 0:
            while total_events == 0:
                click.secho(f"No log events yet, still waiting...\r", fg='yellow', nl=False)
                next_token, num_events = print_logs(logs_client, family, name)
                total_events += num_events
                if (time.time() - start) > 60 * 5:
                    click.secho(
                        f"\nTimeout after 5 minutes, please run the `numerai node status`"
                        f"command for this model or visit the log console:\n"
                        f"https://console.aws.amazon.com/cloudwatch/home?"
                        f"region=us-east-1#logsV2:log-groups/log-group/$252Ffargate$252Fservice$252F{node}"
                        f"/log-events/{name.replace('/', '$252F')}", fg='red'
                    )
                    break

        return


# TODO: harden source of cluster arn
def get_recent_task_status_aws(ecs_client, node, verbose):
    tasks = ecs_client.list_tasks(
        cluster='numerai-submission',
        family=node)

    # try to find stopped tasks
    if len(tasks["taskArns"]) == 0:
        tasks = ecs_client.list_tasks(
            cluster='numerai-submission',
            desiredStatus='STOPPED',
            family=node)

    if len(tasks["taskArns"]) == 0:
        click.secho(f"No recent tasks found...", fg='red')
        return None

    tasks = ecs_client.describe_tasks(
        cluster='numerai-submission',
        tasks=tasks["taskArns"]
    )
    return tasks['tasks'][0]


def get_name_and_print_logs(logs_client, family, limit, next_token=None, raise_on_error=True):
    streams = logs_client.describe_log_streams(
        logGroupName=family,
        orderBy="LastEventTime",
        descending=True)

    if len(streams['logStreams']) == 0:
        if not raise_on_error:
            return False
        raise exception_with_msg(
            "No logs found. Make sure the webhook has triggered by checking "
            "`numerai node status` and make sure a task is in the RUNNING state "
            "(this can take a few minutes). Also, make sure your webhook has "
            "triggered at least once by running `numerai node test`")

    name = streams['logStreams'][0]['logStreamName']
    print_logs(logs_client, family, name, limit, next_token)
    return True


def print_logs(logs_client, family, name, limit=None, next_token=None):
    kwargs = {}  # boto is weird, and doesn't allow `None` for parameters
    if next_token is not None:
        kwargs['nextToken'] = next_token
    if limit is not None:
        kwargs['limit'] = limit

    events = logs_client.get_log_events(
        logGroupName=family,
        logStreamName=name,
        **kwargs
    )

    if len(events["events"]) == limit:
        click.echo('...more log lines available: use -n option to get more...')
    for event in events["events"]:
        click.echo(f"[{name}] {str(datetime.fromtimestamp(event['timestamp'] / 1000))}: {event['message']}")

    return events['nextForwardToken'], len(events['events'])


@click.command()
@click.option('--verbose', '-v', is_flag=True)
@click.option(
    '--num-lines', '-n', type=int, default=20,
    help="the number of log lines to return")
@click.option(
    "--log-type", "-l", type=click.Choice(LOG_TYPES), default=LOG_TYPE_CLUSTER,
    help=f"The log type to lookup. One of {LOG_TYPES}. Default is {LOG_TYPE_CLUSTER}.")
@click.option(
    "--follow-tail", "-f", is_flag=True,
    help="tail the logs constantly")
@click.pass_context
def status(ctx, verbose, num_lines, log_type, follow_tail):
    """
    Get the logs from the latest task for this node

    Logs are not created until a task is in the RUNNING state,
    so the logs returned by this command might be out of date.
    """
    ctx.ensure_object(dict)
    model = ctx.obj['model']
    node = model['name']
    monitor(node, load_or_init_nodes(node), verbose, num_lines, log_type, follow_tail)



import json
import shutil

import click

from numerai.cli.constants import *


def load_config(path):
    if os.stat(path).st_size == 0:
        return {}
    with open(path) as f:
        return json.load(f)


def store_config(path, obj):
    with open(path, 'w+') as f:
        json.dump(obj, f, indent=2)


def maybe_create(path, protected=False):
    created = False

    directory = os.path.dirname(path)
    if not os.path.exists(directory):
        os.makedirs(directory)

    if not os.path.exists(path):
        created = True
        if protected:
            store_config(os.open(path, os.O_CREAT | os.O_WRONLY, 0o600), {})
            os.chmod(path, 0o600)
        else:
            store_config(path, {})

    return created


def load_or_init_nodes(node=None):
    maybe_create(NODES_PATH)
    cfg = load_config(NODES_PATH)
    try:
        return cfg[node] if node else cfg
    except KeyError:
        click.secho(
            'Node has not been configured, run `numerai node --help` '
            'to learn how to create one', fg='red'
        )
        exit(1)


def copy_files(src, dst, force=False, verbose=True):
    if not os.path.exists(dst):
        os.mkdir(dst)
    for filename in os.listdir(src):
        src_file = os.path.join(src, filename)
        dst_file = os.path.join(dst, filename)
        if os.path.exists(dst_file) and not force:
            if not click.confirm(f'{filename} already exists. Overwrite?'):
                return
        if os.path.isdir(src_file):
            if verbose:
                click.secho(f"copying directory {dst_file}", fg='yellow')
            os.makedirs(dst_file, exist_ok=True)
            copy_files(src_file, dst_file, force=force, verbose=verbose)
        else:
            if verbose:
                click.secho(f"copying file {dst_file}", fg='yellow')
            shutil.copy(src_file, dst_file)


def copy_example(example, dest, verbose):
    example_dir = os.path.join(EXAMPLE_PATH, example)
    dst_dir = dest if dest is not None else example

    if Path.home() == dst_dir:
        click.secho("Do not copy example files to your home directory.", fg='red')
        exit(1)

    click.echo(f'Copying {example} example to {dst_dir}')
    copy_files(example_dir, dst_dir, force=False, verbose=verbose)

    dockerignore_path = os.path.join(dst_dir, '.dockerignore')
    if not os.path.exists(dockerignore_path):
        with open(dockerignore_path, 'a+') as f:
            f.write(".numerai\n")
            f.write("numerai_dataset*\n")
            f.write(".git\n")
            f.write("venv\n")
    return dst_dir



import json

import click
from numerapi import base_api

from numerai.cli.constants import *
from numerai.cli.node.config import config
from numerai.cli.node.deploy import deploy
from numerai.cli.node.destroy import destroy
from numerai.cli.node.test import test, status
from numerai.cli.util.keys import get_numerai_keys


@click.group()
@click.option(
    '--model-name', '-m', type=str, prompt=True,
    help=f"The name of one of your models to configure the Prediction Node for."
         f"It defaults to the first model returned from your account."
)
@click.option(
    '--signals', '-s', is_flag=True,
    help=f"Target a signals model with this name. Defaults to false."
)
@click.pass_context
def node(ctx, model_name, signals):
    """
    Commands to manage and test Prediction Nodes.
    """
    if not os.path.exists(CONFIG_PATH):
        click.secho('cannot find .numerai config directory, '
                    'run `numerai setup`', fg='red')
        exit(1)

    if signals:
        tournament = TOURNAMENT_SIGNALS
        name_prefix = 'signals'
    else:
        tournament = TOURNAMENT_NUMERAI
        name_prefix = 'numerai'
    napi = base_api.Api(*get_numerai_keys())
    models = napi.get_models(tournament)

    try:
        model_id = models[model_name]
        ctx.ensure_object(dict)
        ctx.obj['model'] = {
            'id': model_id,
            'name': f'{name_prefix}-{model_name}',
            'is_signals': signals
        }

    except KeyError:
        click.secho(
            f'No tournament {tournament} model with name "{model_name}" '
            f'found in list of models:\n{json.dumps(models, indent=2)}'
            f'\n(use the "-s" flag for signals models)',
            fg='red'
        )
        exit(1)


node.add_command(config)
node.add_command(deploy)
node.add_command(destroy)
node.add_command(test)
node.add_command(status)



import os
import platform
import sys
import json

import click


def exception_with_msg(msg):
    return click.ClickException(msg)


def is_win8():
    if sys.platform == 'win32':
        return '8' in platform.win32_ver()[0]
    return False


def is_win10():
    if sys.platform == 'win32':
        return '10' in platform.win32_ver()[0]
    return False


def is_win10_professional():
    if is_win10():
        # for windows 10 only, we need to know if it's pro vs home
        import winreg
        with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows NT\CurrentVersion") as key:
            return winreg.QueryValueEx(key, "EditionID")[0] == 'Professional'

    return False


# error checking for docker; sadly this is a mess,
# especially b/c there's tons of ways to mess up your docker install
# especially on windows :(
def root_cause(std_out, err_msg):
    all = f'{std_out.decode("utf-8") }\n{err_msg.decode("utf-8") }'
    if b'is not recognized as an internal or external command' in err_msg:
        if sys.platform == 'win32':
            if is_win10_professional():
                raise exception_with_msg(
                    f"Docker does not appear to be installed. Make sure to download/install docker from "
                    f"https://hub.docker.com/editions/community/docker-ce-desktop-windows \n"
                    f"If you're sure docker is already installed,  then for some reason it isn't in your PATH like expected. "
                    f"Restarting may fix it.")

            else:
                raise exception_with_msg(
                    f"Docker does not appear to be installed. Make sure to download/install docker from "
                    f"https://github.com/docker/toolbox/releases and run 'Docker Quickstart Terminal' when you're done."
                    f"\nIf you're sure docker is already installed, then for some reason it isn't in your PATH like expected. "
                    f"Restarting may fix it.")

    if b'command not found' in err_msg:
        if sys.platform == 'darwin':
            raise exception_with_msg(
                f"Docker does not appear to be installed. You can install it with `brew cask install docker` or "
                f"from https://hub.docker.com/editions/community/docker-ce-desktop-mac")

        else:
            raise exception_with_msg(
                f"docker command not found. Please install docker "
                f"and make sure that the `docker` command is in your $PATH")

    if b'This error may also indicate that the docker daemon is not running' in err_msg or b'Is the docker daemon running' in err_msg:
        if sys.platform == 'darwin':
            raise exception_with_msg(
                f"Docker daemon not running. Make sure you've started "
                f"'Docker Desktop' and then run this command again.")

        elif sys.platform == 'linux2':
            raise exception_with_msg(
                f"Docker daemon not running or this user cannot acccess the docker socket. "
                f"Make sure docker is running and that your user has permissions to run docker. "
                f"On most systems, you can add your user to the docker group like so: "
                f"`sudo groupadd docker; sudo usermod -aG docker $USER` and then restarting your computer.")

        elif sys.platform == 'win32':
            if 'DOCKER_TOOLBOX_INSTALL_PATH' in os.environ:
                raise exception_with_msg(
                    f"Docker daemon not running. Make sure you've started "
                    f"'Docker Quickstart Terminal' and then run this command again.")

            else:
                raise exception_with_msg(
                    f"Docker daemon not running. Make sure you've started "
                    f"'Docker Desktop' and then run this command again.")

    if b'invalid mode: /opt/plan' in err_msg:
        if sys.platform == 'win32':
            raise exception_with_msg(
                f"You're running Docker Toolbox, but you're not using the 'Docker Quickstart Terminal'. "
                f"Please re-run `numerai setup` from that terminal.")

    if b'Drive has not been shared' in err_msg:
        raise exception_with_msg(
            f"You're running from a directory that isn't shared to your docker Daemon. "
            f"Make sure your directory is shared through Docker Desktop: "
            f"https://docs.docker.com/docker-for-windows/#shared-drives")

    if b'No configuration files' in err_msg:
        raise exception_with_msg(
            "You're running from a directory that isn't shared to your docker Daemon. \
            Try running from a directory under your HOME, e.g. C:\\Users\\$YOUR_NAME\\$ANY_FOLDER"
        )

    if b'returned non-zero exit status 137' in err_msg:
        raise exception_with_msg(
            "Your docker container ran out of memory. Please open the docker desktop UI"
            " and increase the memory allowance in the advanced settings."
        )

    if b'Temporary failure in name resolution' in err_msg:
        raise exception_with_msg("You network failed temporarily, please try again.")

    if b'No Fargate configuration exists for given values.' in std_out:
        raise exception_with_msg("Invalid size preset, report this to Numerai")

    if 'Can\'t add file' in all or b'Error processing tar file(exit status 1): unexpected EOF' in err_msg:
        err_files = [f for f in all.split('\n') if 'Can\'t add file' in f]
        raise exception_with_msg(
            "Docker was unable to access some files while trying to build,"
            "either another program is using them or docker does not have permissions"
            f"to access them: {json.dumps(err_files, indent=2)}"
        )

    if b'PermissionError: [Errno 13] Permission denied: \'modules.json\'' in err_msg:
        raise exception_with_msg(
            "It looks like Docker daemon is running as root, please restart in rootless"
            "mode: https://docs.docker.com/engine/security/rootless/"
        )

    # these are non-errors that either shouldn't be handled or are handled elsewhere
    if b'Can\'t update submission after deadline' in err_msg:
        return
    if b'ResourceNotFoundException' in std_out or b'NoSuchEntity' in std_out:
        return

    raise exception_with_msg(
        f'Numerai CLI was unable to identify an error, please try to use the '
        f'"--verbose|-v" option for more information before reporting this\n{all}'
    )



import json

import click
from numerapi import base_api

from numerai.cli.constants import *
from numerai.cli.util.docker import terraform, check_for_dockerfile
from numerai.cli.util.files import \
    load_or_init_nodes, \
    store_config, \
    copy_example
from numerai.cli.util.keys import \
    get_provider_keys, \
    get_numerai_keys


@click.command()
@click.option('--verbose', '-v', is_flag=True)
@click.option(
    '--provider', '-P', type=str,
    help=f"Select a cloud provider. One of {PROVIDERS}. "
         f"Defaults to {DEFAULT_PROVIDER}.")
@click.option(
    '--size', '-s', type=str,
    help=f"CPU credits (cores * 1024) and Memory (in MiB) used in the deployed container. "
         f"Defaults to {DEFAULT_SIZE} (run `numerai list-constants` to see options).")
@click.option(
    '--path', '-p', type=str,
    help=f"Target a file path. Defaults to current directory ({DEFAULT_PATH}).")
@click.option(
    '--example', '-e', type=click.Choice(EXAMPLES),
    help=f'Specify an example to use for this node. Options are {EXAMPLES}.')
@click.option(
    '--cron', '-c', type=str,
    help=f'A cron expression to trigger this node on a schedule '
         f'(e.g. "30 18 ? * 7 *" to execute at 18:30 UTC every Saturday). '
         f'This prevents your webhook from auto-registering. '
         f'Check the AWS docs for more info about cron expressions: '
         f'https://docs.aws.amazon.com/AmazonCloudWatch/latest/events/ScheduledEvents.html')
@click.option(
    '--register-webhook', '-r', is_flag=True,
    help=f'Forces your webhook to register with Numerai. '
         f'Use in conjunction with options that prevent webhook auto-registering.')
@click.pass_context
def config(ctx, verbose, provider, size, path, example, cron, register_webhook):
    """
    Uses Terraform to create a full Numerai Compute cluster in AWS.
    Prompts for your AWS and Numerai API keys on first run, caches them in $HOME/.numerai.

    At the end of running, this will output a config file 'nodes.json'.
    """
    ctx.ensure_object(dict)
    model = ctx.obj['model']
    node = model['name']
    model_id = model['id']

    if example is not None:
        path = copy_example(example, path, verbose)

    # get nodes config object and set defaults for this node
    click.secho(f'configuring node "{node}"...')
    nodes_config = load_or_init_nodes()
    nodes_config.setdefault(node, {})
    nodes_config[node].update({
        key: default
        for key, default in DEFAULT_SETTINGS.items()
        if key not in nodes_config[node]
    })
    # update node as needed
    node_conf = nodes_config[node]
    if provider:
        node_conf['provider'] = provider
    if size:
        node_conf['cpu'] = SIZE_PRESETS[size][0]
        node_conf['memory'] = SIZE_PRESETS[size][1]
    if path:
        node_conf['path'] = os.path.abspath(path)
    if model_id:
        node_conf['model_id'] = model_id
    if cron:
        node_conf['cron'] = cron
    nodes_config[node] = node_conf

    # double check there is a dockerfile in the path we are about to configure
    check_for_dockerfile(nodes_config[node]['path'])
    store_config(NODES_PATH, nodes_config)

    # terraform apply
    provider_keys = get_provider_keys(node)
    click.secho(f'running terraform to provision cloud infrastructure...')
    terraform(f'apply -auto-approve', verbose,
              env_vars=provider_keys,
              inputs={'node_config_file': 'nodes.json'})
    click.secho('cloud resources created successfully', fg='green')

    # terraform output for AWS nodes
    click.echo(f'saving node configuration to {NODES_PATH}...')
    res = terraform(f"output -json aws_nodes", verbose).decode('utf-8')
    try:
        aws_nodes = json.loads(res)
    except json.JSONDecodeError:
        click.secho("failed to save node configuration, pleas retry.", fg='red')
        return
    for node_name, data in aws_nodes.items():
        nodes_config[node_name].update(data)
    store_config(NODES_PATH, nodes_config)
    if verbose:
        click.secho(f'new config:\n{json.dumps(load_or_init_nodes(), indent=2)}')

    webhook_url = nodes_config[node]['webhook_url']
    napi = base_api.Api(*get_numerai_keys())
    if not cron or register_webhook:
        click.echo(f'registering webhook {webhook_url} for model {model_id}...')
        napi.set_submission_webhook(model_id, webhook_url)

    else:
        click.echo(f'removing registered webhook for model {model_id}...')
        napi.set_submission_webhook(model_id, None)

    click.secho('Prediction Node configured successfully. '
                'Next: deploy and test your node', fg='green')


""" An extra entry point specifically for training. Used when running locally """

import predict

train_data_path, predict_data_path, predict_output_path = predict.download_data()

model_id = predict.MODEL_ID
model_type = predict.MODEL

predict.train(train_data_path, model_id, model_type, force_training=True)



from setuptools import setup
from os import path

# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='numerai-cli',
    version='0.3.4.dev00',
    description='A library for deploying Numer.ai Prediction Nodes.',
    url='https://github.com/numerai/numerai-cli',
    author='Numer.ai',
    author_email='contact@numer.ai',
    license='MIT',
    packages=['numerai'],
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.6.0',
    install_requires=[
        "click>=7",
        "boto3",
        "numerapi>=2.4.5",
        "colorama",
        "requests",
    ],
    entry_points={
        'console_scripts': ['numerai=numerai:main'],
    }
)



""" Sample tournament model in python 3 """

import os
import logging
import joblib
import numerapi
import pandas as pd
from sklearn.linear_model import LinearRegression

TRAINED_MODEL_PREFIX = './trained_model'

# Pull model id from "MODEL_ID" environment variable
# defaults to None, change to a model id from
MODEL_ID = os.getenv('MODEL_ID', None)
MODEL = LinearRegression()

# initialize API client
napi = numerapi.NumerAPI()


def download_data():
    """ Download the data files """

    logging.info('downloading tournament data files')
    # create an API client and download current data
    file_path = napi.download_current_dataset()
    # we only want the unzipped directory
    file_path = file_path.split('.zip')[0]
    logging.info('output path: %s', file_path)

    train_data_path = f'{file_path}/numerai_training_data.csv'
    predict_data_path = f'{file_path}/numerai_tournament_data.csv'
    predict_output_path = f'{file_path}/predictions.csv'

    return train_data_path, predict_data_path, predict_output_path


def train(train_data_path, model_id, model, force_training=False):
    """ Train the model """

    model_name = TRAINED_MODEL_PREFIX
    if model_id:
        model_name += f"_{model_id}"

    # load a model if we have a trained model already and we aren't forcing a training session
    if os.path.exists(model_name) and not force_training:
        logging.info('loading existing trained model')
        model = joblib.load(model_name)
        return model

    logging.info('reading training data')
    train_data = pd.read_csv(train_data_path)

    logging.info('extracting features and targets')
    # get all rows and all columns from 4th to last-1
    # pylint: disable=no-member
    train_features = train_data.iloc[:, 3:-1]
    # get all rows and only last column
    # pylint: disable=no-member
    train_targets = train_data.iloc[:, -1]
    del train_data

    logging.info('training model')
    model.fit(X=train_features, y=train_targets)

    logging.info('saving model')
    joblib.dump(model, model_name)

    return model


def predict(model, predict_data_path):
    """ Predict the results based on the previously trained model """

    logging.info('reading prediction data')
    predict_data = pd.read_csv(predict_data_path)

    logging.info('extracting features')
    # get all rows and only first column
    # pylint: disable=no-member
    predict_ids = predict_data.iloc[:, 0]
    # get all rows and all columns from 4th to last-1
    # pylint: disable=no-member
    predict_features = predict_data.iloc[:, 3:-1]
    del predict_data

    logging.info('generating predictions')
    predictions = model.predict(predict_features)
    predictions = pd.DataFrame(predictions, columns=['prediction'])
    predictions.insert(0, "id", predict_ids)

    return predictions


def submit(predictions, predict_output_path, model_id=None):
    """ Submit a prediction """

    logging.info('writing predictions to file')
    # numerai doesn't want the index, so don't write it to our file
    predictions.to_csv(predict_output_path, index=False)

    # Numerai API uses Environment variables to find your keys:
    # NUMERAI_PUBLIC_ID and NUMERAI_SECRET_KEY
    # these are set by docker via the numerai cli; see README for more info
    logging.info('submitting')
    napi.upload_predictions(predict_output_path, model_id=model_id)


def main():
    """ Download, train, predict and submit for this model """

    train_data_path, predict_data_path, predict_output_path = download_data()
    trained_model = train(train_data_path, MODEL_ID, MODEL)
    predictions = predict(trained_model, predict_data_path)
    submit(predictions, predict_output_path, MODEL_ID)


if __name__ == '__main__':
    main()
